/*

Miranda IM: the free IM client for Microsoft* Windows*

Copyright 2000-2004 Miranda ICQ/IM project, 
all portions of this codebase are copyrighted to the people 
listed in contributors.txt.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "commonheaders.h"
#include "../random/plugins/m_plugins.h"

typedef struct {
	MIRANDAHOOK pfnHook;
	HWND hwnd;
	UINT message;
} THookSubscriber;

typedef struct {
	char name[MAXMODULELABELLENGTH];
	DWORD hookHash;
	int subscriberCount;
	THookSubscriber *subscriber;
} THookList;

typedef struct {
	char name[MAXMODULELABELLENGTH];
	DWORD nameHash;
	MIRANDASERVICE pfnService;
} TServiceList;

typedef struct {
	int hookId;
	HANDLE hDoneEvent;
	WPARAM wParam;
	LPARAM lParam;
	int result;
} THookToMainThreadItem;

typedef struct {
	HANDLE hDoneEvent;
	WPARAM wParam;
	LPARAM lParam;
	int result;
	const char *name;
} TServiceToMainThreadItem;

static THookList *hook;
static TServiceList *service;
static int hookCount,serviceCount;
static CRITICAL_SECTION csHooks,csServices;
static DWORD mainThreadId;
static HANDLE hMainThread;
extern HANDLE hShutdownEvent;

#ifdef _ALPHA_FUSE_	
#include "../../SDK/headers_c/m_fuse.h"
HINSTANCE hFuseCore=NULL;
int (*fuse_ctrl)(DWORD,void*);
FUSE_LINK fuseLink;
#endif // _ALPHA_FUSE_

int LoadSystemModule(void);		// core: m_system.h services
int LoadDatabaseModule(void);	// core: db
int LoadNewPluginsModuleInfos(void); // core: preloading plugins
int LoadNewPluginsModule(void);	// core: N.O. plugins
int LoadNetlibModule(void);		// core: network
int LoadLangPackModule(void);	// core: translation
int LoadFirstRunModule(void);	// core: used for setup
int LoadProtocolsModule(void);	// core: protocol manager
int LoadIgnoreModule(void);		// protocol filter: ignore

int LoadSendRecvUrlModule(void);	//send/recv
int LoadSendRecvEMailModule(void);	//send/recv
int LoadSendRecvAuthModule(void);	//send/recv
int LoadSendRecvFileModule(void);	//send/recv

int LoadContactListModule(void);// ui: clist
int LoadOptionsModule(void);	// ui: options dialog
int LoadFindAddModule(void);	// ui: search/add users
int LoadSkinModule(void);		// ui: skin
int LoadHelpModule(void);		// ui: help stuff
int LoadUserInfoModule(void);	// ui: user info
int LoadHistoryModule(void);	// ui: history viewer
int LoadAwayMsgModule(void);	// ui: setting away messages
int LoadVisibilityModule(void);	// ui: visibility control
int LoadCLUIModule(void);		// ui: CList UI
int LoadPluginOptionsModule(void);	// ui: plugin viewer
int LoadAddContactModule(void);	// ui: authcontrol contacts
int LoadAutoAwayModule(void);	// ui: away
int LoadUserOnlineModule(void);	// ui: online alert
int LoadUtilsModule(void);		// ui: utils (has a few window classes, like HyperLink)
int LoadCLCModule(void);		// window class: CLC control
int LoadButtonModule(void);		// window class: button class
int LoadContactsModule(void);    // random: contact

static int LoadDefaultModules(void)
{
	int *disableDefaultModule;

    //load order is very important for these
	if(LoadSystemModule()) return 1;
	if(LoadLangPackModule()) return 1; // langpack will be a system module in the new order so this is moved 'ere
	if(LoadDatabaseModule()) return 1;
	if(LoadUtilsModule()) return 1;		//order not important for this, but no dependencies and no point in pluginising	
	if(LoadProtocolsModule()) return 1;
	if(LoadSkinModule()) return 1;
    if(LoadNewPluginsModuleInfos()) return 1;
    if(LoadFirstRunModule()) return 1;
	if(LoadOptionsModule()) return 1;
	if(LoadContactListModule()) return 1;
	if(LoadButtonModule()) return 1;
	if(LoadCLCModule()) return 1;
	if(LoadAddContactModule()) return 1;
	if(LoadContactsModule()) return 1;
	if(LoadNewPluginsModule()) return 1;
	//plugin support
	disableDefaultModule=(int*)CallService(MS_PLUGINS_GETDISABLEDEFAULTARRAY,0,0);
	//order becomes less important below here
	if(!disableDefaultModule[DEFMOD_UIFINDADD]) if(LoadFindAddModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UIUSERINFO]) if(LoadUserInfoModule()) return 1;	
	if(!disableDefaultModule[DEFMOD_SRURL]) if(LoadSendRecvUrlModule()) return 1;
	if(!disableDefaultModule[DEFMOD_SREMAIL]) if(LoadSendRecvEMailModule()) return 1;
	if(!disableDefaultModule[DEFMOD_SRAUTH]) if(LoadSendRecvAuthModule()) return 1;
	if(!disableDefaultModule[DEFMOD_SRFILE]) if(LoadSendRecvFileModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UIHELP]) if(LoadHelpModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UIHISTORY]) if(LoadHistoryModule()) return 1;
	if(!disableDefaultModule[DEFMOD_RNDAUTOAWAY]) if(LoadAutoAwayModule()) return 1;
	if(!disableDefaultModule[DEFMOD_RNDUSERONLINE]) if(LoadUserOnlineModule()) return 1;
	if(!disableDefaultModule[DEFMOD_SRAWAY]) if(LoadAwayMsgModule()) return 1;
	if(!disableDefaultModule[DEFMOD_RNDIGNORE]) if(LoadIgnoreModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UIVISIBILITY]) if(LoadVisibilityModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UICLUI]) if(LoadCLUIModule()) return 1;
	if(!disableDefaultModule[DEFMOD_UIPLUGINOPTS]) if(LoadPluginOptionsModule()) return 1;
	if(!disableDefaultModule[DEFMOD_PROTOCOLNETLIB]) if(LoadNetlibModule()) return 1;
	return 0;
}

int InitialiseModularEngine(void)
{
	#ifdef _ALPHA_FUSE_
		hFuseCore=LoadLibrary("fuse.dll");
		if (hFuseCore) {			
			fuse_ctrl=(void*)GetProcAddress(hFuseCore,"fuse_ctrl");
			if (fuse_ctrl) {
				int rc;
				fuseLink.cbSize=sizeof(FUSE_LINK);
				fuse_ctrl(FUSE_INIT,&fuseLink);
				rc=LoadDefaultModules();
				fuse_ctrl(FUSE_DEFMOD,&rc);
				return rc;
			} else {
				FreeLibrary(hFuseCore); hFuseCore=NULL;
			} //if
		}; //if
	#endif
	hookCount=serviceCount=0;
	hook=NULL;
	service=NULL;
	InitializeCriticalSection(&csHooks);
	InitializeCriticalSection(&csServices);

	mainThreadId=GetCurrentThreadId();
	DuplicateHandle(GetCurrentProcess(),GetCurrentThread(),GetCurrentProcess(),&hMainThread,THREAD_SET_CONTEXT,FALSE,0);

	return LoadDefaultModules();

}

void DestroyingModularEngine(void)
{
	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			fuse_ctrl(FUSE_DEATH,NULL);
			return;
		} //if
	#endif
	return;
}

void DestroyModularEngine(void)
{
	int i;
	#ifdef _ALPHA_FUSE_
		if (hFuseCore) {
			if (fuse_ctrl) {
				fuse_ctrl(FUSE_DEINIT,NULL); fuse_ctrl=NULL;
			} //if
			FreeLibrary(hFuseCore); hFuseCore=NULL;
			return;
		} //if
	#endif
	for(i=0;i<hookCount;i++)
		if(hook[i].subscriberCount) free(hook[i].subscriber);
	if(hookCount) free(hook);
	if(serviceCount) free(service);
	DeleteCriticalSection(&csHooks);
	DeleteCriticalSection(&csServices);
	CloseHandle(hMainThread);
}

DWORD NameHashFunction(const char *szStr)
{
#if defined _M_IX86 && !defined _NUMEGA_BC_FINALCHECK
	__asm {		   //this breaks if szStr is empty
		xor  edx,edx
		xor  eax,eax
		mov  esi,szStr
		mov  al,[esi]
		xor  cl,cl
	lph_top:	 //only 4 of 9 instructions in here don't use AL, so optimal pipe use is impossible
		xor  edx,eax
		inc  esi
		xor  eax,eax
		and  cl,31
		mov  al,[esi]
		add  cl,5
		test al,al
		rol  eax,cl		 //rol is u-pipe only, but pairable
		                 //rol doesn't touch z-flag
		jnz  lph_top  //5 clock tick loop. not bad.

		xor  eax,edx
	}
#else
	DWORD hash=0;
	int i;
	int shift=0;
	for(i=0;szStr[i];i++) {
		hash^=szStr[i]<<shift;
		if(shift>24) hash^=(szStr[i]>>(32-shift))&0x7F;
		shift=(shift+5)&0x1F;
	}
	return hash;
#endif
}

///////////////////////////////HOOKS

static int FindHookByName(const char *name)
{
	int i;

	for(i=0;i<hookCount;i++)
		if(!strcmp(hook[i].name,name)) return i;
	return -1;
}

static int FindHookByHashAndName(DWORD Hash, const char *name)
{
	int i;
	for (i=0; i<hookCount; i++)
		if (hook[i].hookHash==Hash) 
		{
			if (!strcmp(hook[i].name, name)) return i;
		}
	return -1;
}

HANDLE CreateHookableEvent(const char *name)
{
	DWORD Hash = NameHashFunction(name);
	HANDLE ret;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.CreateHookableEvent(name);
		} //if
	#endif

	EnterCriticalSection(&csHooks);
	//if(FindHookByName(name)!=-1) {LeaveCriticalSection(&csHooks); return NULL;}
	if (FindHookByHashAndName(Hash, name) != -1) 
	{
		LeaveCriticalSection(&csHooks);
		return NULL;
	}
	hook=(THookList*)realloc(hook,sizeof(THookList)*(hookCount+1));
	strcpy(hook[hookCount].name,name);
	hook[hookCount].hookHash=Hash;
	hook[hookCount].subscriberCount=0;
	hook[hookCount].subscriber=NULL;
	hookCount++;
	ret=(HANDLE)hookCount;
	LeaveCriticalSection(&csHooks);
	return ret;
}

int DestroyHookableEvent(HANDLE hEvent)
{
	int hookId=(int)hEvent-1;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.DestroyHookableEvent(hEvent);
		} //if
	#endif

	EnterCriticalSection(&csHooks);
	if(hookId>=hookCount || hookId<0) {LeaveCriticalSection(&csHooks); return 1;}
	if(hook[hookId].name[0]==0) {LeaveCriticalSection(&csHooks); return 1;}
	hook[hookId].name[0]=0;
	if(hook[hookId].subscriberCount) {
		free(hook[hookId].subscriber);
		hook[hookId].subscriber=NULL;
		hook[hookId].subscriberCount=0;
	}
	LeaveCriticalSection(&csHooks);
	return 0;
}

static int CallHookSubscribers(int hookId,WPARAM wParam,LPARAM lParam)
{
	int i,returnVal=0;

	EnterCriticalSection(&csHooks);
	if(hookId>=hookCount || hookId<0 || hook[hookId].name[0]==0) returnVal=-1;
	else {		
		//NOTE: We've got the critical section while all this lot are called. That's mostly safe, though.
		for(i=0;i<hook[hookId].subscriberCount;i++) {
			if(hook[hookId].subscriber[i].pfnHook!=NULL) {
				if(returnVal=hook[hookId].subscriber[i].pfnHook(wParam,lParam)) break;
			}
			else if(hook[hookId].subscriber[i].hwnd!=NULL) {
				if(returnVal=SendMessage(hook[hookId].subscriber[i].hwnd,hook[hookId].subscriber[i].message,wParam,lParam)) break;
			}//if
		}//for
	}
	LeaveCriticalSection(&csHooks);
	return returnVal;
}

static void CALLBACK HookToMainAPCFunc(DWORD dwParam)
{
	THookToMainThreadItem *item=(THookToMainThreadItem*)dwParam;

	item->result=CallHookSubscribers(item->hookId,item->wParam,item->lParam);
	SetEvent(item->hDoneEvent);
	return;
}

int NotifyEventHooks(HANDLE hEvent,WPARAM wParam,LPARAM lParam)
{

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.NotifyEventHooks(hEvent,wParam,lParam);
		} //if
	#endif

	if(GetCurrentThreadId()!=mainThreadId) {
		THookToMainThreadItem item;

		item.hDoneEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
		item.hookId=(int)hEvent-1;
		item.wParam=wParam;
		item.lParam=lParam;

		QueueUserAPC(HookToMainAPCFunc,hMainThread,(DWORD)&item);
		WaitForSingleObject(item.hDoneEvent,INFINITE);
		CloseHandle(item.hDoneEvent);
		return item.result;
	}
	else
		return CallHookSubscribers((int)hEvent-1,wParam,lParam);
}

HANDLE HookEvent(const char *name,MIRANDAHOOK hookProc)
{
	int hookId;
	HANDLE ret;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.HookEvent(name,hookProc);
		} //if
	#endif

	EnterCriticalSection(&csHooks);
	hookId=FindHookByName(name);
	if(hookId==-1) {
#ifdef _DEBUG
		MessageBox(NULL,"Attempt to hook non-existant event",name,MB_OK);
#endif
		LeaveCriticalSection(&csHooks);
		return NULL;
	}
	hook[hookId].subscriber=(THookSubscriber*)realloc(hook[hookId].subscriber,sizeof(THookSubscriber)*(hook[hookId].subscriberCount+1));
	hook[hookId].subscriber[hook[hookId].subscriberCount].pfnHook=hookProc;
	hook[hookId].subscriber[hook[hookId].subscriberCount].hwnd=NULL;
	hook[hookId].subscriberCount++;
	ret=(HANDLE)((hookId<<16)|hook[hookId].subscriberCount);
	LeaveCriticalSection(&csHooks);
	return ret;
}

HANDLE HookEventMessage(const char *name,HWND hwnd,UINT message)
{
	int hookId;
	HANDLE ret;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.HookEventMessage(name,hwnd,message);
		} //if
	#endif

	EnterCriticalSection(&csHooks);
	hookId=FindHookByName(name);
	if(hookId==-1) {
#ifdef _DEBUG
		MessageBox(NULL,"Attempt to hook non-existant event",name,MB_OK);
#endif
		LeaveCriticalSection(&csHooks);
		return NULL;
	}
	hook[hookId].subscriber=(THookSubscriber*)realloc(hook[hookId].subscriber,sizeof(THookSubscriber)*(hook[hookId].subscriberCount+1));
	hook[hookId].subscriber[hook[hookId].subscriberCount].pfnHook=NULL;
	hook[hookId].subscriber[hook[hookId].subscriberCount].hwnd=hwnd;
	hook[hookId].subscriber[hook[hookId].subscriberCount].message=message;
	hook[hookId].subscriberCount++;
	ret=(HANDLE)((hookId<<16)|hook[hookId].subscriberCount);
	LeaveCriticalSection(&csHooks);
	return ret;
}

int UnhookEvent(HANDLE hHook)
{
	int hookId=(int)hHook>>16;
	int subscriberId=((int)hHook&0xFFFF)-1;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.UnhookEvent(hHook);
		} //if
	#endif

	EnterCriticalSection(&csHooks);
	if(hookId>=hookCount || hookId<0) {LeaveCriticalSection(&csHooks); return 1;}
	if(hook[hookId].name[0]==0) {LeaveCriticalSection(&csHooks); return 1;}
	if(subscriberId>=hook[hookId].subscriberCount || subscriberId<0) {LeaveCriticalSection(&csHooks); return 1;}
	hook[hookId].subscriber[subscriberId].pfnHook=NULL;
	hook[hookId].subscriber[subscriberId].hwnd=NULL;
	while(hook[hookId].subscriberCount && hook[hookId].subscriber[hook[hookId].subscriberCount-1].pfnHook==NULL && hook[hookId].subscriber[hook[hookId].subscriberCount-1].hwnd==NULL)
		hook[hookId].subscriberCount--;
	if (hook[hookId].subscriberCount==0) {
		if(hook[hookId].subscriber) free(hook[hookId].subscriber);	
		hook[hookId].subscriber=NULL;
	}
	LeaveCriticalSection(&csHooks);
	return 0;
}

/////////////////////SERVICES

static __inline TServiceList *FindServiceByHash(DWORD hash)
{
	int first,last,mid;

	if(serviceCount==0) return NULL;
	first=0; last=serviceCount-1;
	if(hash>=service[last].nameHash) {
		if(hash==service[last].nameHash) return &service[last];
		return NULL;
	}
	for(;;) {
	  mid=(first+last)>>1;
	  if(hash>service[mid].nameHash) {
		  if(last-first<=1) break;
		  first=mid;
	  }
	  else if(hash<service[mid].nameHash) {
		  if(last-first<=1) break;
		  last=mid;
	  }
	  else return &service[mid];
	}
	return NULL;
}

static __inline TServiceList *FindServiceByName(const char *name)
{
	return FindServiceByHash(NameHashFunction(name));
}

/* assume critical section csServices is owned */
static __inline int FindHashForService(DWORD hash)
{	
	int first=0,last=serviceCount-1,mid,i;
	if (serviceCount==0) {
		return 0;
	}
	for (;;) {
		mid=(first+last)>>1;
		if (hash>service[mid].nameHash) {
			/* hash is bigger than anything lower than mid, don't
			include anything before mid in the next search */
			if ((last-first) <= 1) break;
			first=mid;
		} else if (hash<service[mid].nameHash) {
			/* hash is lower than anything after mid, don't include
			anything after mid in the next search */
			if ((last-first) <= 1) break;
			last=mid;
		} else {
			//service already exists
			return -1;
		} //if
	} //for
	/* trouble is, we're looking for the nearest match against the sorted
	array, so look in the smallest range between the found
	first and last (looking any closer is more trouble than it's worth) */
	for(i=first;i<serviceCount;i++) {
		if (hash<service[i].nameHash) break;
	}
	return i;
}

HANDLE CreateServiceFunction(const char *name,MIRANDASERVICE serviceProc)
{
	DWORD hash;
	int i;
#ifdef _DEBUG
	if (name==NULL) { 
		MessageBox(0,"Someone tried to create a NULL'd service, see call stack for more info","",0);
		DebugBreak();
		return NULL;
	}
#else
	if (name==NULL) return NULL;
#endif
	hash=NameHashFunction(name);	
	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.CreateServiceFunction(name,serviceProc);
		} //if
	#endif
	EnterCriticalSection(&csServices);
	i=FindHashForService(hash);
	if (i==-1) {
		LeaveCriticalSection(&csServices);
		return NULL;
	}
	service=(TServiceList*)realloc(service,sizeof(TServiceList)*(serviceCount+1));
	MoveMemory(service+i+1,service+i,sizeof(TServiceList)*(serviceCount-i));
	strcpy(service[i].name,name);
	service[i].nameHash=hash;
	service[i].pfnService=serviceProc;
	serviceCount++;
	LeaveCriticalSection(&csServices);
	return (HANDLE)hash;
}

int DestroyServiceFunction(HANDLE hService)
{
	TServiceList *pService;
	int i;

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.DestroyServiceFunction(hService);
		} //if
	#endif

	EnterCriticalSection(&csServices);
	pService=FindServiceByHash((DWORD)hService);
	if(pService==NULL) {LeaveCriticalSection(&csServices); return 1;}
	i=(int)(pService-service);
	MoveMemory(service+i,service+i+1,sizeof(TServiceList)*(--serviceCount-i));
	LeaveCriticalSection(&csServices);
	return 0;
}

int ServiceExists(const char *name)
{
	int ret;
	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.ServiceExists(name);
		} //if
	#endif
	EnterCriticalSection(&csServices);
	ret=FindServiceByName(name)!=NULL;
	LeaveCriticalSection(&csServices);
	return ret;
}

int CallService(const char *name,WPARAM wParam,LPARAM lParam)
{
	TServiceList *pService;
	MIRANDASERVICE pfnService;

#ifdef _DEBUG
	if (name==NULL) {
		MessageBox(0,"Someone tried to CallService(NULL,..) see stack trace for details","",0);
		DebugBreak();
		return CALLSERVICE_NOTFOUND;
	}
#else
	if (name==NULL) return CALLSERVICE_NOTFOUND;
#endif

	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.CallService(name,wParam,lParam);
		} //if
	#endif
	
	EnterCriticalSection(&csServices);
	pService=FindServiceByName(name);
	if(pService==NULL) {
		LeaveCriticalSection(&csServices);
//#ifdef _DEBUG
//		MessageBox(NULL,"Attempt to call non-existant service",name,MB_OK);
//#endif
		return CALLSERVICE_NOTFOUND;
	}
	pfnService=pService->pfnService;
	LeaveCriticalSection(&csServices);
	return ((int (*)(WPARAM,LPARAM))pfnService)(wParam,lParam);
}

static void CALLBACK CallServiceToMainAPCFunc(DWORD dwParam)
{
	TServiceToMainThreadItem *item = (TServiceToMainThreadItem*) dwParam;
	item->result = CallService(item->name, item->wParam, item->lParam);
	SetEvent(item->hDoneEvent);
}

int CallServiceSync(const char *name, WPARAM wParam, LPARAM lParam)
{
	#ifdef _ALPHA_BASE_
		if (hFuseCore && fuse_ctrl)  {
			return fuseLink.CallServiceSync(name,wParam,lParam);
		} //if
	#endif
	if (name==NULL) return CALLSERVICE_NOTFOUND;
	// the service is looked up within the main thread, since the time it takes
	// for the APC queue to clear the service being called maybe removed.
	// even thou it may exists before the call, the critsec can't be locked between calls.
	if (GetCurrentThreadId() != mainThreadId)
	{
		TServiceToMainThreadItem item;
		item.wParam = wParam;
		item.lParam = lParam;
		item.name = name;
		item.hDoneEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
		QueueUserAPC(CallServiceToMainAPCFunc, hMainThread, (DWORD) &item);
		WaitForSingleObject(item.hDoneEvent, INFINITE);
		CloseHandle(item.hDoneEvent);
		return item.result;
	} else {
		return CallService(name, wParam, lParam);
	}
}