/*

Miranda IM: the free IM client for Microsoft* Windows*

Copyright 2000-2004 Miranda ICQ/IM project, 
all portions of this codebase are copyrighted to the people 
listed in contributors.txt.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "../core/commonheaders.h"
#include "database.h"

int ProfileManager(char *szDbDest,int cbDbDest);
int ShouldAutoCreate(void);
int CreateDbHeaders(void);
int CheckDbHeaders(void);
int InitialiseDbHeaders(void);
int InitSettings(void);
void UninitSettings(void);
int InitContacts(void);
void UninitContacts(void);
int InitEvents(void);
void UninitEvents(void);
int InitCrypt(void);
int InitTime(void);
int InitModuleNames(void);
void UninitModuleNames(void);
int InitCache(void);
void UninitCache(void);
int InitIni(void);
void UninitIni(void);

static char szDbPath[MAX_PATH];
HANDLE hDbFile;
CRITICAL_SECTION csDbAccess;
struct DBHeader dbHeader;

//parsecommandline() will set szdbpath to the right database, or fail
static int ParseCommandLine(void)
{
	if(ShouldShowProfileManager()) return ProfileManager(szDbPath,sizeof(szDbPath));
	return GetDefaultProfilePath(szDbPath,sizeof(szDbPath),NULL);
}

static int LoadDatabase(void)
{
	int isNewDb;
	DWORD bytesRead;

	hDbFile=CreateFile(szDbPath,GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,ShouldAutoCreate()?OPEN_ALWAYS:OPEN_EXISTING,0,NULL);
	if(hDbFile==INVALID_HANDLE_VALUE) {
		MessageBox(NULL,Translate("Unable to open the selected profile.  Trying to gain exclusive access to it failed.  It maybe already open by another Miranda IM"),Translate("Miranda IM Profile Database"),MB_OK|MB_ICONINFORMATION);
		return 1;
	}
	isNewDb=GetFileSize(hDbFile,NULL)==0;
	ReadFile(hDbFile,&dbHeader,sizeof(dbHeader),&bytesRead,NULL);
	if(isNewDb) CreateDbHeaders();
	else {
		char *szMsg=NULL;
		switch(CheckDbHeaders()) {
			case 0: //hurrah
				//if (dbHeader.ofsFileEnd<GetFileSize(hDbFile,NULL)) szMsg=Translate("Miranda has detected dubious values from the selected profile, using it may cause untold amount of profile damage, please visit http://www.miranda-im.org for various profile maintenance tools to fix the problems.");
				break;
			case 1:	// bad header
				szMsg=Translate("Unable to verify the structure of the selected profile.  DBTool will not help!");
				break;
			case 2:	// bad version
				szMsg=Translate("Unable to use the selected profile.  The version information within the profile suggests it's newer than I currently support.");
			case 3:	// bad user
				szMsg=Translate("Unable to use the selected profile.  Various important parts of user data is missing.  You may be able to salvage your history and contacts.");
			default:
				szMsg=Translate("Unable to verify the selected profile.  You may wish to try and fix this profile via dbtool.  Visit http://www.miranda-im.org for more information.");
		} //switch
		if (szMsg) { 
			MessageBox(NULL,szMsg,Translate("Miranda IM Profile Database"),MB_OK|MB_ICONERROR);
			return 666;
		} //if
	}
	InitialiseDbHeaders();
	return 0;
}

static void UnloadDatabase(void)
{
	CloseHandle(hDbFile);
}

DWORD CreateNewSpace(int bytes)
{
	DWORD ofsNew;

	ofsNew=dbHeader.ofsFileEnd;
	dbHeader.ofsFileEnd+=bytes;
	DBWrite(0,&dbHeader,sizeof(dbHeader));
	log2("newspace %d@%08x",bytes,ofsNew);
	return ofsNew;
}

void DeleteSpace(DWORD ofs,int bytes)
{
	PBYTE buf;
	log2("deletespace %d@%08x",bytes,ofs);
	dbHeader.slackSpace+=bytes;
	DBWrite(0,&dbHeader,sizeof(dbHeader));
	buf=(PBYTE)calloc(bytes,1);
	DBWrite(ofs,buf,bytes);
	free(buf);
}

void UnloadDatabaseModule(void)
{
	UninitIni();
	UninitEvents();
	UninitSettings();
	UninitContacts();
	UninitModuleNames();
	UninitCache();
	UnloadDatabase();
	DeleteCriticalSection(&csDbAccess);
}

static int GetProfileName(WPARAM wParam,LPARAM lParam)
{
  char *pszFilename,*pszExt;
  int len;

  pszFilename=strrchr(szDbPath,'\\');
  if(pszFilename==NULL) return 1;
  pszFilename++;
  pszExt=strrchr(pszFilename,'.');
  len=max((int)wParam-1,pszExt-pszFilename);
  strncpy((char*)lParam,pszFilename,len);
  ((char*)lParam)[len]=0;
  return 0;
}

static int GetProfilePath(WPARAM wParam,LPARAM lParam) {
	char *pszTmp = strrchr(szDbPath,'\\');
	if (pszTmp==NULL) return 1;
	strncpy((char*)lParam,szDbPath,pszTmp-szDbPath);
	((char*)lParam)[pszTmp-szDbPath]=0;
	return 0;
}

int LoadDatabaseModule(void)
{
	InitializeCriticalSection(&csDbAccess);
	log0("DB logging running");
	if(ParseCommandLine()) return 1;
	if(LoadDatabase()) return 1;
	if(InitCache()) return 1;
	if(InitModuleNames()) return 1;
	if(InitContacts()) return 1;
	if(InitSettings()) return 1;
	if(InitEvents()) return 1;
	if(InitCrypt()) return 1;
	if(InitTime()) return 1;
	if(InitIni()) return 1;
	CreateServiceFunction(MS_DB_GETPROFILENAME,GetProfileName);
	CreateServiceFunction(MS_DB_GETPROFILEPATH,GetProfilePath);
	return 0;
}

static DWORD DatabaseCorrupted=0;

void __cdecl dbpanic(void *arg)
{
	
	MessageBox(0,Translate("Miranda has detected corruption in your database. This corruption maybe fixed by DBTool.  Please download it from http://www.miranda-im.org. Miranda will now shutdown."),Translate("Database Panic"),MB_SETFOREGROUND|MB_TOPMOST|MB_APPLMODAL|MB_ICONWARNING|MB_OK);
	TerminateProcess(GetCurrentProcess(),255);
	return;
}

void DatabaseCorruption(void)
{
	int kill=0;

	EnterCriticalSection(&csDbAccess);
	if (DatabaseCorrupted==0) {
		DatabaseCorrupted++;
		kill++;
	} else {
		/* db is already corrupted, someone else is dealing with it, wait here
		so that we don't do any more damage */
		LeaveCriticalSection(&csDbAccess);
		Sleep(INFINITE);
		return;
	}
	LeaveCriticalSection(&csDbAccess);
	if (kill) {
		_beginthread(dbpanic,0,NULL);
		Sleep(INFINITE);
	}
}

#ifdef DBLOGGING
void DBLog(const char *file,int line,const char *fmt,...)
{
	FILE *fp;
	va_list vararg;
	char str[1024];

	va_start(vararg,fmt);
	_vsnprintf(str,sizeof(str),fmt,vararg);
	va_end(vararg);
	fp=fopen("c:\\mirandadatabase.log.txt","at");
	fprintf(fp,"%u: %s %d: %s\n",GetTickCount(),file,line,str);
	fclose(fp);
}
#endif